/*
 * Copyright 2001-2003 The eXo platform SARL All rights reserved.
 * Please look at license.txt in info directory for more license detail. 
 */

package org.exoplatform.services.workflow.impl;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.jar.JarInputStream;
import java.util.zip.ZipInputStream;

import org.exoplatform.container.ExoContainer;
import org.exoplatform.container.component.ComponentLifecycle;
import org.exoplatform.container.configuration.ConfigurationManager;
import org.exoplatform.container.xml.InitParams;
import org.exoplatform.services.database.HibernateService;
import org.exoplatform.services.organization.Group;
import org.exoplatform.services.organization.Membership;
import org.exoplatform.services.organization.OrganizationService;
import org.exoplatform.services.workflow.Process;
import org.exoplatform.services.workflow.ProcessInstance;
import org.exoplatform.services.workflow.Task;
import org.exoplatform.services.workflow.Timer;
import org.exoplatform.services.workflow.WorkflowServiceContainer;
import org.hibernate.cfg.Configuration;
import org.jbpm.context.exe.ContextInstance;
import org.jbpm.db.GraphSession;
import org.jbpm.db.JbpmSchema;
import org.jbpm.db.JbpmSession;
import org.jbpm.db.JbpmSessionFactory;
import org.jbpm.graph.def.ProcessDefinition;
import org.jbpm.jpdl.par.ProcessArchiveDeployer;
import org.jbpm.taskmgmt.exe.TaskInstance;

/**
 * Created y the eXo platform team User: Benjamin Mestrallet Date: 28 juin 2004
 */
public class WorkflowServiceContainerImpl implements WorkflowServiceContainer,
    ComponentLifecycle {

  private ProcessesConfig config_;
  private ConfigurationManager configurationManager_;
  private JbpmSessionFactory sessionFactory_;
  private OrganizationService orgService_;
  private ThreadLocal threadLocal_;
  private String hibernateServiceName_;

  public WorkflowServiceContainerImpl(OrganizationService orgService,
      ConfigurationManager conf, InitParams params) throws Exception {
    hibernateServiceName_ = params.getValueParam("hibernate.service")
        .getValue();
    config_ = (ProcessesConfig) params.getObjectParamValues(
        ProcessesConfig.class).get(0);
    this.configurationManager_ = conf;
    threadLocal_ = new ThreadLocal();
    orgService_ = orgService;

  }

  public void initComponent(ExoContainer container) throws Exception {
    HibernateService hservice = null;
    if (hibernateServiceName_ == null || hibernateServiceName_.length() == 0
        || "default".equals(hibernateServiceName_)) {
      hservice = (HibernateService) container
          .getComponentInstanceOfType(HibernateService.class);
    } else {
      hservice = (HibernateService) container
          .getComponentInstance(hibernateServiceName_);
    }
    Configuration hconf = hservice.getHibernateConfiguration();
    sessionFactory_ = JbpmSessionFactory.buildJbpmSessionFactory(hconf);
    JbpmSchema schema = sessionFactory_.getJbpmSchema();
    if (!schema.hasJbpmTables()) {
      schema.createSchema();
    }
    init();
  }

  public void startComponent(ExoContainer arg0) throws Exception {
  }

  public void stopComponent(ExoContainer arg0) throws Exception {
  }

  public void destroyComponent(ExoContainer arg0) throws Exception {
  }

  public JbpmSessionFactory getJbpmSessionFactory() {
    return sessionFactory_;
  }

  public JbpmSession openSession() {
    JbpmSession currentSession = (JbpmSession) threadLocal_.get();
    if (currentSession == null) {
      currentSession = this.sessionFactory_
          .openJbpmSessionAndBeginTransaction();
      threadLocal_.set(currentSession);
    }
    return currentSession;
  }

  private void init() {
    // init the default processes
    JbpmSession session = openSession();
    try {
      if (!session.getGraphSession().findAllProcessDefinitions().isEmpty())
        return;

      HashSet predefinedProcesses = config_.getPredefinedProcess();
      String processLoc = config_.getProcessLocation();

      for (Iterator iter = predefinedProcesses.iterator(); iter.hasNext();) {
        String parFile = (String) iter.next();
        InputStream iS;
        try {
          iS = configurationManager_.getInputStream(processLoc + parFile);
          ProcessArchiveDeployer.deployZipInputStream(new ZipInputStream(iS),
              this.sessionFactory_);
        } catch (Exception e) {
          // process does not exist
        }
      }
    } finally {
      closeSession();
    }
  }

  public void closeSession(JbpmSession session) {
    if (session == null)
      return;
    try {
      session.commitTransactionAndClose();
    } catch (Throwable t) {
    }
    threadLocal_.set(null);
  }

  public void closeSession() {
    JbpmSession s = (JbpmSession) threadLocal_.get();
    if (s != null)
      s.commitTransactionAndClose();
    threadLocal_.set(null);
  }

  public void rollback() {
    JbpmSession s = (JbpmSession) threadLocal_.get();
    if (s != null)
      s.rollbackTransactionAndClose();
    threadLocal_.set(null);
  }

  public void rollback(JbpmSession session) {
    if (session == null)
      return;
    try {
      session.rollbackTransactionAndClose();
    } catch (Throwable t) {
    }
    threadLocal_.set(null);
  }

  public List<Task> getAllTasks(String user) throws Exception {
    List allTasks = new ArrayList();
    allTasks.addAll(getUserTaskList(user));
    allTasks.addAll(getGroupTaskList(user));
    return allTasks;
  }

  public List<Task> getUserTaskList(String user) {
    JbpmSession session = openSession();
    List taskInstances = session.getTaskMgmtSession().findTaskInstances(user);
    return wrapTasks(taskInstances);

  }

  public List<Task> getGroupTaskList(String user) throws Exception {
    List<Task> groupTasks = new ArrayList<Task>();
    String key = null;
    Collection groups = orgService_.getGroupHandler().findGroupsOfUser(user);
    JbpmSession session = openSession();
    for (Iterator iter = groups.iterator(); iter.hasNext();) {
      Group group = (Group) iter.next();
      Collection memberships = 
        orgService_.getMembershipHandler().findMembershipsByUserAndGroup(user, group.getId());
      for (Iterator iterator = memberships.iterator(); iterator.hasNext();) {
        Membership membership = (Membership) iterator.next();
        key = membership.getMembershipType() + ACTOR_ID_KEY_SEPARATOR
            + group.getId();
        List tasks = session.getTaskMgmtSession().findTaskInstances(key);
        if (tasks.size() > 0) {
          groupTasks.addAll(wrapTasks(tasks));
        }
      }
      // manage the wildcard memberships
      key = "*" + ACTOR_ID_KEY_SEPARATOR + group.getId();
      List tasks = session.getTaskMgmtSession().findTaskInstances(key);
      if (tasks.size() > 0) {
        groupTasks.addAll(wrapTasks(tasks));
      }
    }
    return groupTasks;
  }

  private static List<Task> wrapTasks(List tasks) {
    List<Task> wrappedTokens = new ArrayList<Task>();
    for (Iterator iter = tasks.iterator(); iter.hasNext();) {
      TaskInstance task = (TaskInstance) iter.next();
      wrappedTokens.add(new TaskData(task));
    }
    return wrappedTokens;
  }

  public List<Process> getProcesses() {
    List<Process> processes = new ArrayList<Process>();
    JbpmSession session = openSession();
    List definitions = session.getGraphSession().findAllProcessDefinitions();
    for (Iterator iter = definitions.iterator(); iter.hasNext();) {
      ProcessDefinition def = (ProcessDefinition) iter.next();
      processes.add(new ProcessData(def));
    }
    return processes;
  }

  public boolean hasStartTask(String processId) {
    JbpmSession session = openSession();
    ProcessDefinition processDef = session.getGraphSession()
        .loadProcessDefinition(Long.parseLong(processId));
    if (processDef.getTaskMgmtDefinition().getStartTask() != null)
      return true;
    return false;
  }

  public void startProcess(String processId) {
    JbpmSession session = openSession();
    ProcessDefinition processDef = session.getGraphSession()
        .loadProcessDefinition(Long.parseLong(processId));
    org.jbpm.graph.exe.ProcessInstance processInstance = new org.jbpm.graph.exe.ProcessInstance(
        processDef);
    processInstance.signal();
    session.getGraphSession().saveProcessInstance(processInstance);
  }

  public void startProcess(String remoteUser, String processId, Map variables) {
    JbpmSession session = openSession();
    GraphSession graphSession = session.getGraphSession();
    ProcessDefinition processDef = graphSession.loadProcessDefinition(Long
        .parseLong(processId));
    org.jbpm.graph.exe.ProcessInstance instance = new org.jbpm.graph.exe.ProcessInstance(
        processDef);
    ContextInstance contextInstance = instance.getContextInstance();

    variables.put("initiator", remoteUser);
    contextInstance.addVariables(variables);

    TaskInstance taskInstance = instance.getTaskMgmtInstance()
        .createStartTaskInstance();
    taskInstance.setActorId(remoteUser);
    taskInstance.end();

    // instance.signal();
    graphSession.saveProcessInstance(instance);
  }
  
  public void startProcessFromName(String remoteUser, String processName,
      Map variables) {
    JbpmSession session = openSession();
    GraphSession graphSession = session.getGraphSession();
    ProcessDefinition processDef = graphSession.findLatestProcessDefinition(processName);
    System.out.println("PROCESS DEF FOUND "+processDef);
    org.jbpm.graph.exe.ProcessInstance instance = new org.jbpm.graph.exe.ProcessInstance(processDef);
    ContextInstance contextInstance = instance.getContextInstance();
    contextInstance.addVariables(variables);
    TaskInstance taskInstance = instance.getTaskMgmtInstance()
        .createStartTaskInstance();
    taskInstance.setActorId(remoteUser);
    taskInstance.end();
    graphSession.saveProcessInstance(instance);
  }

  public List<ProcessInstance> getProcessInstances(String processId) {
    List<ProcessInstance> processInstances = new ArrayList<ProcessInstance>();
    JbpmSession session = openSession();
    List jbpmProcessInstances = session.getGraphSession().findProcessInstances(
        new Long(processId).longValue());
    for (Iterator iter = jbpmProcessInstances.iterator(); iter.hasNext();) {
      org.jbpm.graph.exe.ProcessInstance processInstance = (org.jbpm.graph.exe.ProcessInstance) iter
          .next();
      processInstances.add(new ProcessInstanceData(processInstance));

    }
    return processInstances;
  }

  public List<Task> getTasks(String processInstanceId) {
    List<Task> tasks = new ArrayList<Task>();
    JbpmSession session = openSession();
    org.jbpm.graph.exe.ProcessInstance processInstance = session
        .getGraphSession().loadProcessInstance(
            new Long(processInstanceId).longValue());

    Collection taskInstances = processInstance.getTaskMgmtInstance()
        .getTaskInstances();
    for (Iterator iter = taskInstances.iterator(); iter.hasNext();) {
      TaskInstance task = (TaskInstance) iter.next();
      tasks.add(new TaskData(task));
    }
    return tasks;
  }

  public Process getProcess(String processId) {
    JbpmSession session = openSession();
    ProcessDefinition processDef = session.getGraphSession()
        .loadProcessDefinition(Long.parseLong(processId));
    return new ProcessData(processDef);
  }

  public Task getTask(String taskId) {
    JbpmSession session = openSession();
    TaskInstance task = session.getTaskMgmtSession().loadTaskInstance(
        Long.parseLong(taskId));
    return new TaskData(task);
  }

  public ProcessInstance getProcessInstance(String processInstance) {
    JbpmSession session = openSession();
    org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = session
        .getGraphSession().loadProcessInstance(Long.parseLong(processInstance));
    return new ProcessInstanceData(jbpmProcessInstance);
  }

  public Map getVariables(String processInstanceId) {
    JbpmSession session = openSession();
    org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = session
        .getGraphSession().loadProcessInstance(
            Long.parseLong(processInstanceId));
    return jbpmProcessInstance.getContextInstance().getVariables();
  }

  public void endTask(String taskId, Map variables) {
    JbpmSession session = openSession();
    TaskInstance taskInstance = session.getTaskMgmtSession().loadTaskInstance(
        Long.parseLong(taskId));
    org.jbpm.graph.exe.ProcessInstance processInstance = taskInstance
        .getToken().getProcessInstance();
    ContextInstance contextInstance = processInstance.getContextInstance();
    contextInstance.addVariables(variables);
    taskInstance.end();
    session.getGraphSession().saveProcessInstance(processInstance);
  }

  public void endTask(String taskId, Map variables, String transition) {
    JbpmSession session = openSession();
    TaskInstance taskInstance = session.getTaskMgmtSession().loadTaskInstance(
        Long.parseLong(taskId));
    org.jbpm.graph.exe.ProcessInstance processInstance = taskInstance
        .getToken().getProcessInstance();
    ContextInstance contextInstance = processInstance.getContextInstance();
    contextInstance.addVariables(variables);
    taskInstance.end(transition);
    session.getGraphSession().saveProcessInstance(processInstance);
  }

  public List<Timer> getTimers() {
    List<Timer> timers = new ArrayList<Timer>();
    JbpmSession session = openSession();
    Iterator jbpmTimers = session.getSchedulerSession().findTimersByDueDate();
    while (jbpmTimers.hasNext()) {
      org.jbpm.scheduler.exe.Timer jbpmTimer = (org.jbpm.scheduler.exe.Timer) jbpmTimers
          .next();
      timers.add(new TimerData(jbpmTimer));
    }
    return timers;
  }

  public void deployProcess(InputStream iS) throws IOException {
    JarInputStream jarIs = new JarInputStream(iS);
    JbpmSession session = openSession();
    session.getContextSession();
    ProcessArchiveDeployer.deployZipInputStream(new ZipInputStream(jarIs),
        session.getJbpmSessionFactory());
  }

}