/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package javax.jcr;

/**
 * The entry point into the content repository.
 * Represents the entry point into the content repository. Typically the object
 * implementing this interface will be acquired from a JNDI-compatible
 * naming and directory service.
 */
public interface Repository {

    /**
     * The descriptor key for the version of the specification
     * that this repository implements.
     */
     public static final String SPEC_VERSION_DESC = "jcr.specification.version";

    /**
     * The descriptor key for the name of the specification
     * that this repository implements.
     */
     public static final String SPEC_NAME_DESC = "jcr.specification.name";

    /**
     * The descriptor key for the name of the repository vendor.
     */
     public static final String REP_VENDOR_DESC = "jcr.repository.vendor";

    /**
     * The descriptor key for the URL of the repository vendor.
     */
     public static final String REP_VENDOR_URL_DESC = "jcr.repository.vendor.url";

    /**
     * The descriptor key for the name of this repository implementation.
     */
     public static final String REP_NAME_DESC = "jcr.repository.name";

    /**
     * The descriptor key for the version of this repository implementation.
     */
     public static final String REP_VERSION_DESC = "jcr.repository.version";

    /**
     * The presence of this key indicates that this implementation supports
     * all level 1 features. This key will always be present.
     */
     public static final String LEVEL_1_SUPPORTED = "level.1.supported";

    /**
     * The presence of this key indicates that this implementation supports
     * all level 2 features.
     */
     public static final String LEVEL_2_SUPPORTED = "level.2.supported";

    /**
     * The presence of this key indicates that this implementation supports transactions.
     */
     public static final String OPTION_TRANSACTIONS_SUPPORTED = "option.transactions.supported";

    /**
     * The presence of this key indicates that this implementation supports versioning.
     */
     public static final String OPTION_VERSIONING_SUPPORTED = "option.versioning.supported";

    /**
     * The presence of this key indicates that this implementation supports observation.
     */
     public static final String OPTION_OBSERVATION_SUPPORTED = "option.observation.supported";

    /**
     * The presence of this key indicates that this implementation supports locking.
     */
     public static final String OPTION_LOCKING_SUPPORTED = "option.locking.supported";

    /**
     * The presence of this key indicates that this implementation supports the SQL query language.
     */
     public static final String OPTION_QUERY_SQL_SUPPORTED = "option.query.sql.supported";

    /**
     * The presence of this key indicates that the index position notation for
     * same-name siblings is supported within XPath queries.
     */
     public static final String QUERY_XPATH_POS_INDEX = "query.xpath.pos.index";

    /**
     * The presence of this key indicates that XPath queries return results in document order.
     */
     public static final String QUERY_XPATH_DOC_ORDER = "query.xpath.doc.order";

    /**
     * Returns a string array holding all descriptor keys available for this implementation.
     * This set must contain at least the built-in keys defined by the string constants in
     * this interface.Used in conjunction with {@link #getDescriptor(String name)}
     * to query information about this repository implementation.
     */
    public String[] getDescriptorKeys();

    /**
     * Returns the descriptor for the specified key. Used to query information about this
     * repository implementation. The set of available keys can be found by calling
     * {@link #getDescriptorKeys}. If the specifed key is not found, <code>null</code> is returned.
     *
     * @param key a string corresponding to a descriptor for this repository implementation.
     * @return a descriptor string
     */
    public String getDescriptor(String key);

    /**
     * Authenticates the user using the supplied <code>credentials</code>.
     * <p>
     * If <code>workspaceName</code> is recognized as the name of an existing workspace in the repository and
     * authorization to access that workspace is granted, then a new <code>Session</code> object is returned.
     * The format of the string <code>workspaceName</code> depends upon the implementation.
     * <p>
     * If <code>credentials</code> is <code>null</code>, it is assumed that authentication is handled by a
     * mechanism external to the repository itself (for example, through the JAAS framework) and that the
     * repository implementation exists within a context (for example, an application server) that allows
     * it to handle authorization of the request for access to the specified workspace.
     * <p>
     * If <code>workspaceName</code> is <code>null</code>, a default workspace is automatically selected by
     * the repository implementation. This may, for example, be the "home workspace" of the user whose
     * credentials were passed, though this is entirely up to the configuration and implementation of the
     * repository. Alternatively, it may be a "null workspace" that serves only to provide the method
     * {@link Workspace#getAccessibleWorkspaceNames}, allowing the client to select from among available "real"
     * workspaces.
     * <p>
     * If authentication or authorization for the specified workspace fails, a <code>LoginException</code> is
     * thrown.
     * <p>
     * If <code>workspaceName</code> is not recognized, a <code>NoSuchWorkspaceException</code> is thrown.
     *
     * @param credentials   The credentials of the user
     * @param workspaceName the name of a workspace.
     * @return a valid session for the user to access the repository.
     * @throws LoginException  If the login fails.
     * @throws NoSuchWorkspaceException If the specified <code>workspaceName</code> is not recognized.
     * @throws RepositoryException if another error occurs.
     */
    public Session login(Credentials credentials, String workspaceName) throws LoginException, NoSuchWorkspaceException, RepositoryException;

    /**
     * Equivalent to <code>login(credentials, null)</code>.
     *
     * @param credentials   The credentials of the user
     * @return a valid session for the user to access the repository.
     * @throws LoginException  If the login authentication fails.
     * @throws RepositoryException if another error occurs.
     */
    public Session login(Credentials credentials) throws LoginException, RepositoryException;

    /**
     * Equivalent to <code>login(null, workspaceName)</code>.
     *
     * @param workspaceName the name of a workspace.
     * @return a valid session for the user to access the repository.
     * @throws LoginException  If the login authentication fails.
     * @throws NoSuchWorkspaceException If the specified <code>workspaceName</code> is not recognized.
     * @throws RepositoryException if another error occurs.
     */
    public Session login(String workspaceName) throws LoginException, NoSuchWorkspaceException, RepositoryException;

    /**
     * Equivalent to <code>login(null, null)</code>.
     *
     * @return a valid session for the user to access the repository.
     * @throws LoginException  If the login authentication fails.
     * @throws RepositoryException if another error occurs.
     */
    public Session login() throws LoginException, RepositoryException;
}
