/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.ReferentialIntegrityException;
import javax.jcr.RepositoryException;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.datamodel.IllegalPathException;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.storage.cache.WorkspaceCache;
import org.exoplatform.services.jcr.storage.WorkspaceDataContainer;
import org.exoplatform.services.log.LogService;

/**
 * Created by The eXo Platform SARL .<br>
 * Workspace-level data manager
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov
 *         </a>
 * @version $Id: WorkspaceDataManager.java 3565 2006-01-03 13:13:53Z geaz $
 */

public class WorkspaceDataManager {

  protected WorkspaceDataContainer dataContainer;
  
  protected WorkspaceCache cache;

  protected Log log;
  
  protected List listeners;
  
  protected HashMap savedItems;
  
  protected LogService logService;
  
  public WorkspaceDataManager(WorkspaceDataContainer dataContainer,
      WorkspaceCache cache, 
      LogService logService) {
    this.logService = logService;
    log = logService.getLog("org.exoplatform.services.jcr");
    this.dataContainer = dataContainer;
    this.cache = cache;
    this.listeners = new ArrayList();
    
    //listeners.add(cache);
    addItemPersistenceListener(cache);
    
    savedItems = new HashMap();
  }
  
  public ItemData getItemData(JCRPath path) throws RepositoryException {
    // 2. Try from cache
    ItemData data = cache.get(path.getInternalPath());

    // 3. Try from container
    if (data == null) {
      dataContainer.beginTransaction();
      data = (ItemData) dataContainer.getItem(path.getInternalPath());
      dataContainer.commit();
      // no such data
      if (data == null)
        return null;
      else {
        cache.put(data);
      }
    }

    return data;
  }

  public ItemData getItemDataByUUID(String uuid) throws RepositoryException {
    // 2. Try from cache
    ItemData data = cache.get(uuid);

    // 3. Try from container
    if (data == null) {
        dataContainer.beginTransaction();
        data = (ItemData) dataContainer.getItemByUUID(uuid);
        dataContainer.commit();
      // no such data
      if (data == null)
        return null;
      else {
        cache.put(data);
      }
    }
    
    return data;

  }

  public List getReferences(String uuid) throws RepositoryException {
    try {
      InternalQPath versionStorage = InternalQPath.parse(Constants.JCR_VERSION_STORAGE_URI);

      dataContainer.beginTransaction();
      List refProps = dataContainer.getReferences(uuid);
      for(int i=0; i<refProps.size(); i++) {
        PropertyData ref = (PropertyData)refProps.get(i);
        // version (frozen data) never included
        // TODO move to versionable WDM
        if(ref.getQPath().isDescendantOf(versionStorage, false)) {
//        if(isVSDescendant(ref.getQPath())) {
          refProps.remove(i);
        }
      }
      return refProps;
    } catch (IllegalPathException e){
      // never thrown
      throw new RepositoryException("Illegal Path "+e);
    } finally {
      dataContainer.commit();
    }
  }

  public synchronized void delete(ItemImpl item) throws RepositoryException, 
  InvalidItemStateException {
    
    dataContainer.beginTransaction();
    
//    List items = new ArrayList();
//    traverseChangedItems(item.getActualItemData(), items, ItemChangeState.DELETED);

    doDelete(item);
    
    dataContainer.commit();

    // notify
    notifySaveItem(item, ItemChangeState.DELETED);

    log.debug("DataManager COMMITTED deleted " + item.getPath()
        + " to persist. ");

  }

  
  public synchronized void doDelete(ItemImpl item) throws RepositoryException, 
  InvalidItemStateException {
    // check once again as another session may update the same
    //ItemData sameItem = dataContainer.getItemByUUID(item.getActualItemData().getUUID());
    // TODO
    //if(sameItem == null) 
      //return;
      //  throw new InvalidItemStateException("(delete) Item "+item.getPath()+" not found. Probably was deleted by another session");
    //else if(sameItem.getPersistedVersion() > item.getActualItemData().getPersistedVersion())
    //  throw new InvalidItemStateException("(delete) Item "+item.getPath()+" has stale data (version: "+
    //      item.getActualItemData().getPersistedVersion()+" less than stored: "+sameItem.getPersistedVersion()+" ). Probably the data was changed by another session");
    

    dataContainer.delete(item.getActualItemData());
    
    item.setPersistedVersion(item.getPersistedVersion()+1);
  }
  
  public synchronized void update(ItemImpl item) throws RepositoryException, 
  InvalidItemStateException {
    
    dataContainer.beginTransaction();
    
    doUpdate(item);
    
    dataContainer.commit();

    // notify
    notifySaveItem(item, ItemChangeState.UPDATED);


  }
  
  public synchronized void doUpdate(ItemImpl item) throws RepositoryException, 
  InvalidItemStateException {
    // check once again as another session may update the same
    ItemData sameItem = dataContainer.getItemByUUID(item.getActualItemData().getUUID());
    if(sameItem == null)
      throw new InvalidItemStateException("(update) Item "+item.getPath()+" not found. Probably was deleted by another session");
    // TODO
    //else if(sameItem.getPersistedVersion() > item.getActualItemData().getPersistedVersion())
    //  throw new InvalidItemStateException("(update) Item "+item.getPath()+" has stale data (version: "+
    //      item.getActualItemData().getPersistedVersion()+" less than stored: "+sameItem.getPersistedVersion()+" ). Probably the data was changed by another session");
    
    ItemData data;
    if (item.isNode())
      data = dataContainer.update((NodeData) item.getActualItemData());
    else
      data = dataContainer.update((PropertyData) item.getActualItemData());
    
    item.setPersistedVersion(data.getPersistedVersion());

    log.debug("WorkspaceDataManager COMMITTED updated " + item
        + " to persist. version "+data.getPersistedVersion()+" "+item.getPersistedVersion());

  }

  public synchronized void add(ItemImpl item) throws RepositoryException, 
  InvalidItemStateException {
    
    dataContainer.beginTransaction();

    doAdd(item);
    
    dataContainer.commit();
    
    // notify
    notifySaveItem(item, ItemChangeState.ADDED);

  }
  
  protected void doAdd(ItemImpl item) throws RepositoryException, 
  InvalidItemStateException {
    // check once again as another session may insert the same
    ItemData sameItem = dataContainer.getItem(item.getLocation().getInternalPath());
    
    if(sameItem != null && sameItem.isNode() == item.isNode()) {
      throw new InvalidItemStateException("Item "+item.getPath()+
          " (persisted version: "+sameItem.getPersistedVersion()+") already exists in "+
          dataContainer.getName()+". Probably was added by another session "+sameItem.getUUID());
    }
    
    ItemData data;
    //System.out.println("WorkspaceDataManager COMMITTING added " + item.getPath()
    //    + " to persist. "+dataContainer.getName());
    
    if (item.isNode()) {
      data = dataContainer.add((NodeData) item.getActualItemData());
    } else {
      data = dataContainer.add((PropertyData) item.getActualItemData());
    }

    setAddedPersistedVersion(item, data.getPersistedVersion());
  }

  public Calendar getCurrentTime() {
    return dataContainer.getCurrentTime();
  }
  
  public void validate(ItemImpl item) throws RepositoryException,
  AccessDeniedException, ReferentialIntegrityException, LockException,
  ConstraintViolationException{
    
    // Check permissions
    // delete item
    if (item.getState() == ItemChangeState.DELETED) {
      
      // check for manadatory property
      if(!item.isNode()) {
        PropertyImpl prop = (PropertyImpl) item;
        if(prop.getDefinition().isMandatory())
          throw new ConstraintViolationException("Can not remove mandatory property "+prop.getPath());
      }
      
      if (!item.hasPermission(PermissionType.REMOVE))
        throw new AccessDeniedException("Access denied for REMOVE "
            + item.getPath());
      
    }
    
    // add node
    if (item.isNode() && item.getState() == ItemChangeState.ADDED) {

      // check for manadatory property
      if(item.isNode()) {
        NodeImpl node = (NodeImpl) item;
        node.validateMandatoryChildren();
      }

      if(!item.hasPermission(PermissionType.ADD_NODE)) 
        throw new AccessDeniedException("Access denied for ADD "
          + item.getPath());
      
    }
    
    // add/update property
    if (!item.isNode()
        && (item.getState() == ItemChangeState.UPDATED 
            || item.getState() == ItemChangeState.ADDED)) {
      
      if(!item.hasPermission(PermissionType.SET_PROPERTY)) 
        throw new AccessDeniedException("Access denied for ADD "
          + item.getPath());
      
    }

  }
  
  /**
   * @return Returns the dataContainer assotiated with this manager.
   */
  public WorkspaceDataContainer getDataContainer() {
    return dataContainer;
  }
  

  
  // -----------------------------------------------------

  
  private void setAddedPersistedVersion(ItemImpl item, int version) throws RepositoryException {
    item.setPersistedVersion(version);
    if (item.isNode()) {
      NodeImpl node = (NodeImpl) item;
      List props = node.loadChildProperties();
      for (int i = 0; i < props.size(); i++) {
        setAddedPersistedVersion((PropertyImpl) props.get(i), version);
      }
      List nodes = node.loadChildNodes();
      for (int i = 0; i < nodes.size(); i++) {
        setAddedPersistedVersion((NodeImpl) nodes.get(i), version);
      }
    }
  }


  // ---------------------------------------------
  
  public void addItemPersistenceListener(ItemsPersistenceListener listener) {
    listeners.add(listener);
    log.info("Workspace Data manager of '"+this.dataContainer.getName()+"' registered listener: "+listener);
  }

  protected void notifySaveItem(ItemImpl item, int state) throws RepositoryException {

    List items = new ArrayList();
    SessionImpl session = (SessionImpl)item.getSession();
    if(state == ItemChangeState.UPDATED) {
      items.add(new ItemState(item.getActualItemData(), ItemChangeState.UPDATED));
      notifySaveItems(items, session);
      
    } else if(state == ItemChangeState.ADDED) {
      traverseChangedItems(item.getActualItemData(), items, ItemChangeState.ADDED);
      notifySaveItems(items, session);

    } else if(state == ItemChangeState.DELETED) {
      traverseChangedItems(item.getActualItemData(), items, ItemChangeState.DELETED);
      notifySaveItems(items, session);
    }
    
    if(savedItems.containsKey(item.getSession())) {
      ((List)savedItems.get(item.getSession())).addAll(items); 
    }

  }  

  
  protected void notifySaveItems(List itemStates, SessionImpl session) {
    for(Iterator i=listeners.iterator(); i.hasNext();) {
      ItemsPersistenceListener listener = (ItemsPersistenceListener) i.next();
      listener.onSaveItems(itemStates, session);
    }
  }

  protected void traverseChangedItems(ItemData item, List items, int state) throws RepositoryException {
    items.add(new ItemState(item, state));
    //log.debug("WS DATA NODE >> "+item.getQPath().getAsString()+" "+state);
    if (item.isNode()) {
      int newState = state;
      if(newState == ItemChangeState.ADDED)
        newState = ItemChangeState.AUTO_ADDED;
      else if(newState == ItemChangeState.DELETED)
        newState = ItemChangeState.AUTO_DELETED;
      NodeData node = (NodeData) item;
      List props = node.getChildProperties();
      for (int i = 0; i < props.size(); i++) {
        traverseChangedItems((PropertyData) props.get(i), items, newState);
      }
      List nodes = node.getChildNodes();
      for (int i = 0; i < nodes.size(); i++) {
        traverseChangedItems((NodeData) nodes.get(i), items, newState);
      }
    }
  }
  
}